@@ -16,56 +16,47 @@ import os, subprocess, shlex
 from pathlib import Path
 import typing as T
 
-from . import ExtensionModule, ModuleReturnValue
+from . import ExtensionModule, ModuleReturnValue, ModuleState, NewExtensionModule
 from .. import mlog, build
 from ..mesonlib import (MesonException, Popen_safe, MachineChoice,
-                       get_variable_regex, do_replacement)
-from ..interpreterbase import InterpreterObject, InterpreterException, FeatureNew
-from ..interpreterbase import stringArgs, permittedKwargs
-from ..interpreter import Interpreter, DependencyHolder, InstallDirHolder
+                       get_variable_regex, do_replacement, extract_as_list)
+from ..interpreterbase import InterpreterException, FeatureNew
+from ..interpreterbase import permittedKwargs, typed_pos_args
 from ..compilers.compilers import CFLAGS_MAPPING, CEXE_MAPPING
-from ..dependencies.base import InternalDependency, PkgConfigDependency
-from ..environment import Environment
+from ..dependencies import InternalDependency, PkgConfigDependency
 from ..mesonlib import OptionKey
 
-class ExternalProject(InterpreterObject):
+class ExternalProject(NewExtensionModule):
     def __init__(self,
-                 interpreter: Interpreter,
-                 subdir: str,
-                 project_version: T.Dict[str, str],
-                 subproject: str,
-                 environment: Environment,
-                 build_machine: str,
-                 host_machine: str,
-                 configure_command: T.List[str],
+                 state: ModuleState,
+                 configure_command: str,
                  configure_options: T.List[str],
                  cross_configure_options: T.List[str],
                  env: build.EnvironmentVariables,
                  verbose: bool):
-        InterpreterObject.__init__(self)
+        super().__init__()
         self.methods.update({'dependency': self.dependency_method,
                              })
 
-        self.interpreter = interpreter
-        self.subdir = Path(subdir)
-        self.project_version = project_version
-        self.subproject = subproject
-        self.env = environment
-        self.build_machine = build_machine
-        self.host_machine = host_machine
+        self.subdir = Path(state.subdir)
+        self.project_version = state.project_version
+        self.subproject = state.subproject
+        self.env = state.environment
+        self.build_machine = state.build_machine
+        self.host_machine = state.host_machine
         self.configure_command = configure_command
         self.configure_options = configure_options
         self.cross_configure_options = cross_configure_options
         self.verbose = verbose
         self.user_env = env
 
-        self.name = self.subdir.name
         self.src_dir = Path(self.env.get_source_dir(), self.subdir)
         self.build_dir = Path(self.env.get_build_dir(), self.subdir, 'build')
         self.install_dir = Path(self.env.get_build_dir(), self.subdir, 'dist')
         self.prefix = Path(self.env.coredata.get_option(OptionKey('prefix')))
         self.libdir = Path(self.env.coredata.get_option(OptionKey('libdir')))
         self.includedir = Path(self.env.coredata.get_option(OptionKey('includedir')))
+        self.name = self.src_dir.name
 
         # On Windows if the prefix is "c:/foo" and DESTDIR is "c:/bar", `make`
         # will install files into "c:/bar/c:/foo" which is an invalid path.
@@ -76,23 +67,23 @@ class ExternalProject(InterpreterObject):
         # self.prefix is an absolute path, so we cannot append it to another path.
         self.rel_prefix = self.prefix.relative_to(self.prefix.root)
 
-        self.make = self.interpreter.find_program_impl('make')
+        self.make = state.find_program('make')
         self.make = self.make.get_command()[0]
 
-        self._configure()
+        self._configure(state)
 
         self.targets = self._create_targets()
 
-    def _configure(self):
+    def _configure(self, state: ModuleState):
         # Assume it's the name of a script in source dir, like 'configure',
         # 'autogen.sh', etc).
         configure_path = Path(self.src_dir, self.configure_command)
-        configure_prog = self.interpreter.find_program_impl(configure_path.as_posix())
+        configure_prog = state.find_program(configure_path.as_posix())
         configure_cmd = configure_prog.get_command()
 
         d = [('PREFIX', '--prefix=@PREFIX@', self.prefix.as_posix()),
              ('LIBDIR', '--libdir=@PREFIX@/@LIBDIR@', self.libdir.as_posix()),
-             ('INCLUDEDIR', '--includedir=@PREFIX@/@INCLUDEDIR@', self.includedir.as_posix()),
+             ('INCLUDEDIR', None, self.includedir.as_posix()),
              ]
         self._validate_configure_options(d)
 
@@ -140,7 +131,9 @@ class ExternalProject(InterpreterObject):
         # Ensure the user at least try to pass basic info to the build system,
         # like the prefix, libdir, etc.
         for key, default, val in variables:
-            key_format = '@{}@'.format(key)
+            if default is None:
+                continue
+            key_format = f'@{key}@'
             for option in self.configure_options:
                 if key_format in option:
                     break
@@ -160,16 +153,16 @@ class ExternalProject(InterpreterObject):
         if missing:
             var_list = ", ".join(map(repr, sorted(missing)))
             raise EnvironmentException(
-                "Variables {} in configure options are missing.".format(var_list))
+                f"Variables {var_list} in configure options are missing.")
         return out
 
     def _run(self, step: str, command: T.List[str]):
-        mlog.log('External project {}:'.format(self.name), mlog.bold(step))
+        mlog.log(f'External project {self.name}:', mlog.bold(step))
         m = 'Running command ' + str(command) + ' in directory ' + str(self.build_dir) + '\n'
-        log_filename = Path(mlog.log_dir, '{}-{}.log'.format(self.name, step))
+        log_filename = Path(mlog.log_dir, f'{self.name}-{step}.log')
         output = None
         if not self.verbose:
-            output = open(log_filename, 'w')
+            output = open(log_filename, 'w', encoding='utf-8')
             output.write(m + '\n')
             output.flush()
         else:
@@ -178,7 +171,7 @@ class ExternalProject(InterpreterObject):
                                       stderr=subprocess.STDOUT,
                                       stdout=output)
         if p.returncode != 0:
-            m = '{} step returned error code {}.'.format(step, p.returncode)
+            m = f'{step} step returned error code {p.returncode}.'
             if not self.verbose:
                 m += '\nSee logs: ' + str(log_filename)
             raise MesonException(m)
@@ -196,8 +189,8 @@ class ExternalProject(InterpreterObject):
         if self.verbose:
             cmd.append('--verbose')
 
-        target_kwargs = {'output': '{}.stamp'.format(self.name),
-                         'depfile': '{}.d'.format(self.name),
+        target_kwargs = {'output': f'{self.name}.stamp',
+                         'depfile': f'{self.name}.d',
                          'command': cmd + ['@OUTPUT@', '@DEPFILE@'],
                          'console': True,
                          }
@@ -212,16 +205,14 @@ class ExternalProject(InterpreterObject):
                                 install_mode=None,
                                 exclude=None,
                                 strip_directory=True,
-                                from_source_dir=False)
+                                from_source_dir=False,
+                                subproject=self.subproject)
 
-        return [self.target, InstallDirHolder(idir)]
+        return [self.target, idir]
 
-    @stringArgs
     @permittedKwargs({'subdir'})
-    def dependency_method(self, args, kwargs):
-        if len(args) != 1:
-            m = 'ExternalProject.dependency takes exactly 1 positional arguments'
-            raise InterpreterException(m)
+    @typed_pos_args('external_project.dependency', str)
+    def dependency_method(self, state, args: T.Tuple[str], kwargs):
         libname = args[0]
 
         subdir = kwargs.get('subdir', '')
@@ -236,8 +227,8 @@ class ExternalProject(InterpreterObject):
 
         version = self.project_version['version']
         incdir = []
-        compile_args = ['-I{}'.format(abs_includedir)]
-        link_args = ['-L{}'.format(abs_libdir), '-l{}'.format(libname)]
+        compile_args = [f'-I{abs_includedir}']
+        link_args = [f'-L{abs_libdir}', f'-l{libname}']
         libs = []
         libs_whole = []
         sources = self.target
@@ -245,31 +236,27 @@ class ExternalProject(InterpreterObject):
         variables = []
         dep = InternalDependency(version, incdir, compile_args, link_args, libs,
                                  libs_whole, sources, final_deps, variables)
-        return DependencyHolder(dep, self.subproject)
+        return dep
 
 
 class ExternalProjectModule(ExtensionModule):
     @FeatureNew('External build system Module', '0.56.0')
     def __init__(self, interpreter):
         super().__init__(interpreter)
+        self.methods.update({'add_project': self.add_project,
+                             })
 
-    @stringArgs
     @permittedKwargs({'configure_options', 'cross_configure_options', 'verbose', 'env'})
-    def add_project(self, state, args, kwargs):
-        if len(args) != 1:
-            raise InterpreterException('add_project takes exactly one positional argument')
+    @typed_pos_args('external_project_mod.add_project', str)
+    def add_project(self, state: ModuleState, args: T.Tuple[str], kwargs: T.Dict[str, T.Any]):
         configure_command = args[0]
-        configure_options = kwargs.get('configure_options', [])
-        cross_configure_options = kwargs.get('cross_configure_options', ['--host=@HOST@'])
+        configure_options = extract_as_list(kwargs, 'configure_options')
+        cross_configure_options = extract_as_list(kwargs, 'cross_configure_options')
+        if not cross_configure_options:
+            cross_configure_options = ['--host=@HOST@']
         verbose = kwargs.get('verbose', False)
         env = self.interpreter.unpack_env_kwarg(kwargs)
-        project = ExternalProject(self.interpreter,
-                                  state.subdir,
-                                  state.project_version,
-                                  state.subproject,
-                                  state.environment,
-                                  state.build_machine,
-                                  state.host_machine,
+        project = ExternalProject(state,
                                   configure_command,
                                   configure_options,
                                   cross_configure_options,
